# coding: utf-8

# redis的数据类型
REDIS_TYPE = {
    'string': 'string',
    'list': 'list',
    'hash': 'hash',
    'set': 'set',
    'sorted_set': 'zset',
    'not_exists': 'none',
}

# redis列表和有序集的range边界
LEFTMOST = 0
RIGHTMOST = -1



# coding:utf-8

__all__ = ['Mutable', 'Counter']

__metaclass__ = type

import collections
import redis.exceptions as redispy_exception

from ooredis.const import REDIS_TYPE
from ooredis.mix.key import Key

class SingleValue(Key):
    """ 为key对象提供set，get，getset操作。 """

    def __repr__(self):
        key_type = self.__class__.__name__.title()
        key_name = self.name
        key_value = self.get()
        return "{0} Key '{1}': {2}".format(key_type, key_name, key_value)

    def set(self, value, preserve=False, expire=None):
        """ 将key对象的值改为value。

        Args:
            preserve: 指定是否不覆盖存在值。
            expire: 设置key的过期时间，以秒为单位。

        Time:
            O(1)

        Returns:
            None

        Raises:
            TypeError: 当key非空且key不是指定类型时抛出。
            ValueError: 根据preserve参数的情况抛出。
        """
        # set命令可以无视类型进行设置的命令。
        # 为了保证类型的限制，
        # ooredis里对一个非字符串类型进行set将引发TypeError异常。
        if self.exists:
            if self._represent != REDIS_TYPE['string']:
                raise TypeError
            if preserve:
                raise ValueError
        
        value = self._type_case.to_redis(value)
        if expire:
            self._client.setex(self.name, value, expire)
        else:
            self._client.set(self.name, value)

    def get(self):
        """ 返回key对象的值。

        Time:
            O(1)

        Returns:
            value： key的值。
            None: key不存在时返回。

        Raises:
            TypeError: 当key非空且不是string类型时抛出。
        """
        try:
            value = self._client.get(self.name)
            return self._type_case.to_python(value)
        except redispy_exception.ResponseError:
            raise TypeError

    def getset(self, value):
        """ 修改key的值，并返回key之前的值。

        Args:
            value: key对象的新值。

        Time:
            O(1)

        Returns:
            None: 当key不存时(没有前值)返回。
            value: 否则，返回key之前的值。

        Raises:
            TypeError: 当key非空且不是string类型时抛出。
        """
        try:
            value = self._type_case.to_redis(value)
            old = self._client.getset(self.name, value)
            return self._type_case.to_python(old)
        except redispy_exception.ResponseError:
            raise TypeError


class Counter(SingleValue):
    """ 为数值类型的key对象加上incr，decr以及+= 和-= 语法糖。 """

    def incr(self, increment=1):
        """ 将key对象的值加上增量increment。

        Args:
            increment: 增量，默认为1.

        Time:
            O(1)

        Returns:
            int: 操作执行之后的值。

        Raises:
            TypeError: 当key储存的不是数值类型时抛出。
        """
        # redis-py用incr代替redis的incrby
        try:
            return self._client.incr(self.name, increment)
        except redispy_exception.ResponseError:
            raise TypeError

    def decr(self, decrement=1):
        """ 将key对象的值减去减量decrement。

        Args:
            decrement: 减量，默认为1.

        Time:
            O(1)

        Returns:
            int: 操作执行之后的值。

        Raises:
            TypeError: 当key储存的不是数值类型时抛出。
        """
        # redis-py用decr代替redis的decrby
        try:
            return self._client.decr(self.name, decrement)
        except redispy_exception.ResponseError:
            raise TypeError

    def __iadd__(self, increment):
        """ self.incr方法的一个python语法糖。

        和self.incr的区别是，这个特殊方法不返回key值。

        Args:
            increment: 增量。

        Time:
            O(1)

        Returns:
            self: python要求这个特殊方法返回被修改后的对象。

        Raises:
            TypeError: 当key储存的不是数值类型时由self.incr抛出。
        """
        self.incr(increment)
        return self

    def __isub__(self, decrement):
        """ self.decr方法的一个python语法糖。

        和self.decr的区别是，这个特殊方法不返回key的值。

        Args:
            decrement: 减量。

        Time:
            O(1)

        Returns:
            self: python要求这个特殊方法返回被修改后的对象。

        Raises:
            TypeError: 当key储存的不是数值类型时由self.decr抛出。
        """
        self.decr(decrement)
        return self



# coding:utf-8

__all__ = ['SortedSet']

__metaclass__ = type

import redis.exceptions as redispy_exception
from functools import partial

from ooredis.mix.key import Key
from ooredis.const import LEFTMOST, RIGHTMOST

# redis command execute status code
MEMBER_NOT_IN_SET_AND_REMOVE_FALSE = 0
MEMBER_NOT_IN_SET_AND_DELETE_FALSE = 0
MEMBER_NOT_IN_SET_AND_GET_RANK_FALSE = None
MEMBER_NOT_IN_SET_AND_GET_SCORE_FALSE = None

# ZRANGE result item index
VALUE = 0
SCORE= 1

# default value of SortedSet.incr & SortedSet.decr
DEFAULT_INCREMENT = DEFAULT_DECREMENT = 1

class SortedSet(Key):
    """ 有序集对象，底层是redis的zset实现。 """

    def __repr__(self):
        key_type = self.__class__.__name__.title()
        key_name = self.name
        key_values = list(self)
        return "{0} Key '{1}': {2}".format(key_type, key_name, key_values)
    
    def __len__(self):
        """ 返回有序集的基数。
        
        Time:
            O(1)

        Returns:
            int: 有序集的基数，集合不存在或为空时返回0。

        Raises：
            TypeError: 当key不是有序集类型时抛出。
        """
        try:
            return self._client.zcard(self.name)
        except redispy_exception.ResponseError:
            raise TypeError

    def __contains__(self, element):
        """ 检查给定元素是否是有序集的成员。 

        Args:
            element: 要检查的元素。

        Time:
            O(1)

        Returns:
            bool: 如果element是集合的成员，返回True，否则False。

        Raises:
            TypeError: 当key不是有序集类型时抛出。
        """
        # NOTE:
        # 因为在redis里有序集zset没有集合set那样的SISMEMBER命令，
        # 所以这里用ZSCORE命令hack一个，
        # 如果ZSCORE key member不为None，证明element是有序集成员。

        # WARNING:
        # 这里不要用self.score，这两个方法互相引用。
        try:
            element = self._type_case.to_redis(element)
            return None != self._client.zscore(self.name, element)
        except redispy_exception.ResponseError:
            raise TypeError

    def __setitem__(self, member, score):
        """ 将元素的member的score值设为score。

        如果member不存在于有序集，将member加入到有序集。

        Time:
            O(1)

        Returns:
            None

        Raises:
            TypeError: 当key不是有序集类型时抛出。
        """
        try:
            member = self._type_case.to_redis(member)
            self._client.zadd(self.name, member, score) 
        except redispy_exception.ResponseError:
            raise TypeError

    def __getitem__(self, index):
        """ 返回有序集指定下标内的元素。

        Args:
            index: 一个下标或一个slice对象。

        Returns:
            dict： 使用下标时，返回一个字典，包含value和score。
            list：使用slice对象时，返回一个列表，
                  列表中每个项都是一个字典。

        Time:
            O(log(N)+M)，N为有序集的基数，而M为结果集的基数。

        Raises:
            KeyError: index下标超出范围时抛出。
            TypeError: 当key不是有序集类型时抛出。
        """
        try:
            item_to_dict_list = partial(map, lambda item: dict(value=self._type_case.to_python(item[VALUE]), score=item[SCORE]))

            if isinstance(index, slice):
                items = self._client.zrange(self.name, LEFTMOST, RIGHTMOST, withscores=True)
                return item_to_dict_list(items[index])
            else:
                items = self._client.zrange(self.name, index, index, withscores=True)
                return item_to_dict_list(items)[0]
        except redispy_exception.ResponseError:
            raise TypeError

    def __delitem__(self, index):
        """ 删除有序集指定下标内的元素。

        Args:
            index: 下标或一个slice对象。

        Time:
            O(log(N)+M)，N为有序集的基数，而M为被移除成员的数量。

        Returns:
            None

        Raises:
            KeyError: index下标超出范围时抛出。
            TypeError: 当key不是有序集类型时抛出。
        """
        try:
            if isinstance(index, slice):
                start = LEFTMOST if index.start == None else index.start
                stop = RIGHTMOST if index.stop == None else index.stop-1

                self._client.zremrangebyrank(self.name, start, stop)
            else:
                if self._client.zremrangebyrank(self.name, index, index) == \
                   MEMBER_NOT_IN_SET_AND_DELETE_FALSE:
                    raise IndexError
        except redispy_exception.ResponseError:
            raise TypeError

    def remove(self, member, check=False):
        """ 移除有序集成员member，如果member不存在，不做动作。

        Args:
            member: 要移除的成员。
            check: 是否检查member必须在有序集合中。

        Time:
            O(log(N))

        Returns:
            None

        Raises:
            TypeError: 当key不是有序集类型时抛出。
            KeyError: 当member不存在且check为True时抛出。
        """
        try:
            member = self._type_case.to_redis(member)
            status = self._client.zrem(self.name, member)
            if check and status == MEMBER_NOT_IN_SET_AND_REMOVE_FALSE:
                raise KeyError
        except redispy_exception.ResponseError:
            raise TypeError

    def rank(self, member, reverse=False):
        """ 返回有序集中成员member的score值的排名。

        排序可以选择递增(从小到大)和递减(从大到小)两种顺序，
        默认为递增序排序。

        Args:
            member: 被检查的成员。
            reverse: 如果为True，元素以递减排序。

        Time:
            O(log(N))

        Returns:
            int: member的score排名。

        Raises:
            KeyError: 当member不在有序集中时抛出。
            TypeError: 当key不是有序集类型时由in语句抛出。
        """
        try:
            get_rank = self._client.zrevrank if reverse else self._client.zrank

            member = self._type_case.to_redis(member)
            result = get_rank(self.name, member)

            if result == MEMBER_NOT_IN_SET_AND_GET_RANK_FALSE:
                raise KeyError
            else:
                return result
        except redispy_exception.ResponseError:
            raise TypeError

    def score(self, member):
        """ 返回有序集中成员member的score值。

        Args:
            member

        Time:
            O(1)

        Returns:
            float: 以浮点值表示的score值。

        Raises:
            KeyError: 当member不在有序集中时抛出。
            TypeError: 当key不是有序集类型时抛出，由in语句抛出。
        """
        try:
            member = self._type_case.to_redis(member)
            result = self._client.zscore(self.name, member)

            if result == MEMBER_NOT_IN_SET_AND_GET_SCORE_FALSE:
                raise KeyError
            else:
                return result
        except redispy_exception.ResponseError:
            raise TypeError

    def incr(self, member, increment=DEFAULT_INCREMENT):  
        """ 将member的score值加上increment。 

        当key不存在，或member不是key的成员时，
        self.incr(member, increment)等同于self[member] = increment

        Args:
            member: 集合成员
            increment：增量，默认为1。

        Time:
            O(log(N))

        Returns:
            float: member成员的新score值。

        Raises:
            TypeError: 当key不是有序集类型时抛出。
        """
        try:
            member = self._type_case.to_redis(member)
            return self._client.zincrby(self.name, member, increment)
        except redispy_exception.ResponseError:
            raise TypeError

    def decr(self, member, decrement=DEFAULT_DECREMENT):
        """ 将member的score值减去decrement。 

        当key不存在，或member不是key的成员时，
        self.decr(member, decrement)等同于self[member] = 0-decrement

        Args:
            member: 集合成员
            decrement：减量，默认为1。

        Time:
            O(log(N))

        Returns:
            float: member成员的新score值。

        Raises:
            TypeError: 当key不是有序集类型时抛出。
        """
        return self.incr(member, 0-decrement)



# coding:utf-8

__all__ = ['Dict']

__metaclass__ = type

import collections
import redis.exceptions as redispy_exception

from ooredis.const import REDIS_TYPE
from ooredis.mix.key import Key

DEFAULT_INCREMENT = DEFAULT_DECREMENT = 1

class Dict(Key, collections.MutableMapping):
    """ 一个字典对象，底层是redis的hash实现。 """

    def __repr__(self):
        key_type = self.__class__.__name__.title()
        key_name = self.name
        key_values = dict(self.items())
        return "{0} Key '{1}': {2}".format(key_type, key_name, key_values)

    def __setitem__(self, key, value):
        """ 将dict[key]的值设为value。
        如果dict[key]已经存在，则将其覆盖。

        Args:
            key: 字典的键
            value: 字典的值

        Time:
            O(1)

        Returns:
            None

        Raises:
            TypeError: 如果传入的value不是适合的类型。
                       或key对象不是字典类型。
        """
        try:
            value = self._type_case.to_redis(value)
            self._client.hset(self.name, key, value)
        except redispy_exception.ResponseError:
            raise TypeError

    def __getitem__(self, key):
        """ 返回dict[key]的值。
        如果dict[key]不存在，抛出KeyError。

        Args:
            key: 字典的键

        Time:
            O(1)

        Returns:
            value: dict[key]的值

        Raises:
            KeyError: key不存在时抛出。
            TypeError: 如果value不是适合的类型抛出。
                       或key对象不是hash类型。
        """
        # NOTE: 将TypeError的抛出单独抽取出来，
        #       是为了让MIXIN方法的行为和python一致。
        if self.exists and self._represent != REDIS_TYPE['hash']:
            raise TypeError

        # WARNING: 不要在这里使用key in self语句，
        #          除非你自己实现了__contains__方法，
        #          否则迎接你的将是一个无限递归。。。
        if self._client.hexists(self.name, key) is False:
            raise KeyError

        value = self._client.hget(self.name, key)
        return self._type_case.to_python(value)

    def __delitem__(self, key):
        """ 删除dict[key]。
        如果dict[key]不存在，沉默。

        Args:
            key: 字典的键

        Time:
            O(1)

        Returns:
            None

        Raises:
            TypeError: 当key对象不是字典类型时抛出。
        """
        try:
            self._client.hdel(self.name, key)
        except redispy_exception.ResponseError:
            raise TypeError

    def __iter__(self):
        """ 返回字典所有key。

        Time:
            O(N)

        Returns:
            iterator: 包含所有字典所有key的一个生成器。
        
        Raises:
            TypeError: 当key对象不是字典类型时抛出。
        """
        try:
            for key in self._client.hkeys(self.name):
                yield key
        except redispy_exception.ResponseError:
            raise TypeError

    def __len__(self):
        """ 返回字典key-value对的个数，空字典返回0。

        Time:
            O(1)

        Returns:
            int: 字典中key-value对的个数。

        Raises:
            TypeError: 当key对象不是hash类型时抛出。
        """
        try:
            return self._client.hlen(self.name)
        except redispy_exception.ResponseError:
            raise TypeError

    def incr(self, key, increment=DEFAULT_INCREMENT):
        """ 将dict[key]中的值加上increment。

        Args:
            key: 键
            increment：增量

        Time:
            O(1)

        Returns:
            int: 当前dict[key]中的值

        Raises:
            KeyError:如果dict[key]存在且储存的不是整数类型时抛出。
            TypeError:对非字典类型进行incr操作时抛出。
        """
        if self.exists and self._represent != REDIS_TYPE['hash']:
            raise TypeError
        try:
            return self._client.hincrby(self.name, key, increment)
        except redispy_exception.ResponseError:
            raise KeyError

    def decr(self, key, decrement=DEFAULT_DECREMENT):
        """ 将dict[key]中的值减去decrement。

        Args:
            key: 键
            decrement：减量

        Time:
            O(1)

        Returns:
            int: 当前dict[key]中的值

        Raises:
            KeyError:如果dict[key]存在且储存的不是整数类型时抛出。
            TypeError:对非字典类型进行incr操作时抛出。
        """
        return self.incr(key, 0-decrement)



# coding: utf-8

__all__ = [
    'connect', 'get_client',
    'type_case',
    'Dict', 'List', 'Set', 'SortedSet', 'SingleValue', 'Counter',
]

from client import connect, get_client

import type_case

from mix.dict import Dict
from mix.list import List
from mix.set import Set
from mix.sorted_set import SortedSet
from mix.string import SingleValue, Counter



# coding:utf-8

__all__ = ['Key']

__metaclass__ = type

from ooredis.client import get_client
from ooredis.type_case import GenericTypeCase

class Key:
    """ key对象的基类，所有类型的key都继承这个类。 """

    def __init__(self, name, client=None, type_case=None):
        """ 为key对象指定名字和客户端。

        Args:
            name: key的名字
            client: 客户端，默认为全局客户端。
        """
        self._client = client or get_client()
        self._name = name
        self._type_case = type_case or GenericTypeCase

    def __eq__(self, other):
        """ 判断两个key是否相等。

        Args:
            other: 另一个key对象。

        Time:
            O(1)

        Returns:
            bool
        """
        return self.name == other.name

    def __str__(self):
        """ 为key对象创建一个酷酷的打印信息。 """
        key_type = self.__class__.__name__
        return "<{0} Key '{1}'>".format(key_type.title(), self.name)

    @property
    def name(self):
        """ 返回key对象的名字。 """
        return self._name

    @property
    def _represent(self):
        """ 返回key在redis中的表示(实现类型)。

        Time:
            O(1)

        Returns: 
            redis的类型定义在REDIS_TYPE常量中，包含以下：
            'none' : 值不存在
            'string' : 字符串
            'list' : 列表
            'set' : 集合
            'zset' : 有序集
            'hash' : 哈希表
        """
        return self._client.type(self.name)

    @property
    def ttl(self):
        """ 返回key的生存时间。

        Time:
            O(1)

        Returns: 
            None: 值不存在，或值没有设置生存时间。
            long: 以秒为单位的生存时间值。
        """
        # NOTE: py3.0之后整数只有int类型了
        return self._client.ttl(self.name)

    @property
    def exists(self):
        """ 检查key是否存在。

        Time:
            O(1)

        Returns:
            bool: key存在返回True，否则为False。
        """
        return self._client.exists(self.name)

    # TODO: 这个delete只能删除单个key，
    #       考虑增加一个删除多个key的类方法
    def delete(self):
        """ 删除key。


        Time:
            O(1)

        Returns:
            None

        Raises:
            AssertionError: 操作因为异常情况而失败时抛出。
        """
        # redis-py里，对不存在的key进行删除返回False， 删除成功返回True，
        # 这里只删除存在的key，返回None，如果出现问题导致删除失败，抛出异常。
        if self.exists:
            assert(self._client.delete(self.name))

    def expire(self, second):
        """ 为key设置生存时间
        
        Time:
            O(1)

        Returns:
            None

        Raises:
            TypeError: 尝试对不存在的key进行设置时抛出。
            AssertionError: 操作因为异常情况而失败时抛出。
        """
        # redis-py对不存在的key进行expire返回false，
        # 这里抛出一个异常。
        if self.exists == False:
            raise TypeError

        assert(self._client.expire(self.name, second))

    def expireat(self, unix_timestamp):
        """ 为key设置生存时间，以一个unix时间戳为终止时间。

        Time:
            O(1)

        Returns:
            None

        Raises:
            TypeError: 对一个不存在的key进行操作时抛出。
            AssertionError: 操作因为异常情况而失败时抛出。
        """
        # redis-py对不存在的key进行expireat返回False，
        # 这里抛出一个异常。
        if self.exists == False:
            raise TypeError

        assert(self._client.expireat(self.name, unix_timestamp))

    def persist(self):
        """ 移除key的生存时间

        Time:
            O(1)

        Returns:
            None

        Raises:
            TypeError: 对一个不存在的key进行操作时抛出
            AssertionError: 操作因为异常情况而失败时抛出
        """
        # redis-py对不存在的key进行persist返回-1
        # 这里抛出一个异常
        if self.exists == False:
            raise TypeError

        # redis-py对没有生存时间的key进行persist也返回-1
        # 这里选择不进行其他动作(返回None)
        if self.ttl == None:
            return

        assert(self._client.persist(self.name))


# TODO:
class SortableKey:
    """ 为可排序的key提供sort方法。 """

    def sort(self, *args, **kwargs):
        # TODO: ...
        raise NotImplementedError

    def sortget(self, value):
        """ sort方法的一个hack，相当于外键来使用，可以获取多个外键。

        将
        >>> r.sort('user_id', by='not_exists_key', get=('#', 'user_name_*', 'user_password_*') 
        >>> ['222', 'hacker', 'hey,im in', '59230', 'jack', 'jack201022', '2', 'huangz', 'nobodyknows', '1', 'admin', 'a_long_long_password']
        
        改成
        >>> Key('user_id').get('#', 'user_name_*', 'user_password_*')

        或
        >>> Key('user_id').get('#').get('user_name_*').get('user_password_*')

        以及limit
        >>> Key('user_id').get('#').get('user_name_*').get('user_password_*').skip(skip_num).count(count_num)

        列出前10条，key.get() == key.get('#')
        >>> Key('user_id').get().get('user_name_*').get('user_password_*').skip(0).count(10)


        这个方法虽然很cool，但是GET只支持key级别操作，似乎实际效果不是很大。
        """
        pass


# helper

def get_key_name_from_single_value(value):
    """ 从单个值中获取key对象的名字。 """
    return value.name if isinstance(value, Key) else value
   
def get_key_name_from_list(iterable):
    """ 从列表/元组中获取多个key对象的名字。 """
    return map(get_key_name_from_single_value, iterable)



# coding: utf-8

__all__ = ['connect', 'get_client']

__metaclass__ = type

import redis

class SingletonClientMeta(type):
    
    _instance = None

    def __call__(cls, *args, **kwargs):
        if cls._instance == None:
            cls._instance = redis.Redis(*args, **kwargs)
            return cls._instance
        return cls._instance

class RedisClient:
    
    __metaclass__ = SingletonClientMeta


# helper functions:

def connect(*args, **kwargs):
    return RedisClient(*args, **kwargs)

def get_client():
    return RedisClient()



# coding:utf-8

__all__ = ['Set']

__metaclass__ = type

import collections
import redis.exceptions as redispy_exception

from ooredis.mix.key import (
    get_key_name_from_single_value,
    Key,
)

MEMBER_NOT_IN_SET_AND_MOVE_FALSE = 0

def get_members(set_or_set_object):
    """ 从集合对象或集合中提取成员。 """
    return set(set_or_set_object)

class Set(Key):
    """ 集合类型，底层实现是redis的set类型。 """

    def __repr__(self):
        key_type = self.__class__.__name__.title()
        key_name = self.name
        key_values = set(self)
        return "{0} Key '{1}': {2}".format(key_type, key_name, key_values)

    def __len__(self):
        """ 返回集合的基数。

        Time:
            O(1)

        Returns:
            int: 集合的基数。

        Raises:
            TypeError: 当key不是集合类型时抛出。
        """
        try:
            return self._client.scard(self.name)
        except redispy_exception.ResponseError:
            raise TypeError

    def __iter__(self):
        """ 返回集合的可迭代版本。

        Time:
            O(N)

        Returns:
            generator: 一个包含集合所有成员的迭代器。

        Raises:
            TypeError: 当key不是集合类型时抛出。
        """
        try:
            for member in self._client.smembers(self.name):
                yield self._type_case.to_python(member)
        except redispy_exception.ResponseError:
            raise TypeError

    def __contains__(self, element):
        """ 查元素element是否是集合的成员。

        Args:
            element

        Time:
            O(1)

        Returns:
            bool: element是集合成员的话True，否则False。

        Raises:
            TypeError: 当key不是集合类型时抛出。
        """
        try:
            element = self._type_case.to_redis(element)
            return self._client.sismember(self.name, element)
        except redispy_exception.ResponseError:
            raise TypeError

    def add(self, element):
        """ 将element加入到集合当中。

        如果element已经是集合的成员，不做动作。
        如果key不存在，一个空集合被创建并执行add动作。

        Args:
            element: 要加入到集合的元素

        Time:
            O(1)

        Returns:
            None

        Raises:
            TypeError: 当key非空且key不是集合类型时抛出。
        """
        # TODO: redis2.3，add支持多个element加入。
        try:
            element = self._type_case.to_redis(element)
            self._client.sadd(self.name, element)
        except redispy_exception.ResponseError:
            raise TypeError

    def remove(self, element, check=False):
        """ 如果element是集合的成员，移除它。

        如果element不是集合成员且check为False，返回。
        如果element不是集合成员且check为True，抛出KeyError。

        Args：
            element：要删除的成员。
            check：是否检查element是否存在。

        Time：
            O(1)

        Returns:
            None

        Raises:
            TypeError: 当key非空且key不是集合类型时抛出。
            KeyError： check为True且element不是集合成员时抛出。
        """
        try:
            element = self._type_case.to_redis(element)
            delete_success = self._client.srem(self.name, element)
            if not delete_success and check:
                raise KeyError
        except redispy_exception.ResponseError:
            raise TypeError

    def pop(self):
        """ 移除并返回集合中任意一个成员。

        如果集合为空，抛出KeyError。

        Time：
            O(1)

        Returns:
            member: 集合成员。 

        Raises:
            TypeError: 当key非空且key不是集合类型时由抛出。
            KeyError: 集合为空时抛出。
        """
        try:
            element = self._client.spop(self.name)
            if element != None:
                return self._type_case.to_python(element)
            else:
                raise KeyError
        except redispy_exception.ResponseError:
            raise TypeError

    def random(self):
        """ 返回集合中的一个随机元素。

        该操作和pop相似，但pop将随机元素从集合中移除并返回，
        而random则仅仅返回随机元素，而不对集合进行任何改动。

        Time:
            O(1)

        Returns:
            member: 当集合不为空时，返回一个成员。
            None: 当集合为空时，返回None。

        Raises:
            TypeError: 当key非空且key不是集合类型时抛出。
        """
        try:
            element = self._client.srandmember(self.name)
            return self._type_case.to_python(element)
        except redispy_exception.ResponseError:
            raise TypeError

    def move(self, destination, member):
        """ 将集合成员member移动到集合destination中去。

        具体参考Redis命令：SMOVE。

        Args:
            destination: 目的地集合，可以是集合的名字(字符串)，
                         也可以是一个集合key对象。
            member: 要移动的源集合的成员。

        Time:
            O(1)

        Returns:
            None

        Raises:
            KeyError: member不存在于集合时抛出。
            TypeError: 当key非空且key不是集合类型时抛出。
        """
        try:
            destination = get_key_name_from_single_value(destination)
            member = self._type_case.to_redis(member)

            if self._client.smove(self.name, destination, member) == \
               MEMBER_NOT_IN_SET_AND_MOVE_FALSE:
                raise KeyError
        except redispy_exception.ResponseError:
            raise TypeError

    def isdisjoint(self, other):
        """ 检查集合是否和另一个集合不相交。

        Args:
            other: 一个python集合或集合key对象。
        
        Returns:
            bool

        Time:
            O(N)

        Raises:
            TypeError: 当key或other不是指定类型时抛出。
        """
        return get_members(self).isdisjoint(get_members(other))

    def __le__(self, other):
        """ 测试集合是否是另一个集合的子集。

        Args:
            other: 一个python集合或集合key对象。

        Returns:
            bool

        Time:
            O(N)

        Raises:
            TypeError: 当key或other不是指定类型时抛出。
        """
        return get_members(self) <= get_members(other)

    def __lt__(self, other):
        """ 测试集合是否是另一个集合的真子集。

        Args:
            other: 一个python集合或集合key对象。

        Returns:
            bool

        Time:
            O(N)

        Raises:
            TypeError: 当key或other不是指定类型时抛出。
        """
        return get_members(self) < get_members(other)

    def __ge__(self, other):
        """ 测试集合是否是另一个集合的超集。

        Args:
            other: 一个python集合或集合key对象。

        Returns:
            bool

        Time:
            O(N)

        Raises:
            TypeError: 当key或other不是指定类型时抛出。
        """
        return get_members(self) >= get_members(other)

    def __gt__(self, other):
        """ 测试集合是否是另一个集合的真超集。

        Args:
            other: 一个python集合或集合key对象。

        Returns:
            bool

        Time:
            O(N)

        Raises:
            TypeError: 当key或other不是指定类型时抛出。
        """
        return get_members(self) > get_members(other)

    def __or__(self, other):
        """ 返回集合和另一个集合的并集。

        Args:
            other: 一个python集合或集合key对象。

        Time:
            O(N)

        Returns:
            set

        Raises：
            TypeError: 当key或other不是指定类型时抛出。
        """
        return get_members(self) | get_members(other)

    __ror__ = __or__
    __ror__.__doc__ = \
    """ __or__的反向方法，用于支持多集合对象进行并集操作。"""

    def __and__(self, other):
        """ 返回集合和另一个集合的交集。

        Args:
            other: 一个python集合或集合key对象。


        Time:
            O(N)

        Returns:
            set

        Raises：
            TypeError: 当key或other不是指定类型时抛出。
        """
        return get_members(self) & get_members(other)

    __rand__ = __and__
    __rand__.__doc__ = \
    """ __and__的反向方法，用于支持多集合进行交集操作。 """


    def __sub__(self, other):
        """ 返回集合对另一个集合的差集。

        Args:
            other: 一个python集合或集合key对象。

        Time:
            O(N)

        Returns:
            set

        Raises：
            TypeError: 当key或other不是指定类型时抛出。
        """
        return get_members(self) - get_members(other)

    def __rsub__(self, other):
        """ __sub__的反向方法，
        用于支持多集合进行差集运算。
        """
        # WARNING: 注意这里的位置不要弄反。
        return get_members(other) - get_members(self)

    def __xor__(self, other):
        """ 返回集合对另一个集合的对等差集。

        Args:
            other: 一个python集合或集合key对象。

        Returns:
            set

        Time:
            O(N)

        Raises:
            TypeError: 当key或other不是指定类型时抛出。
        """
        return get_members(self) ^ get_members(other)

    __rxor__ = __xor__
    __rxor__.__doc__ = \
    """ __xor__的反向方法，用于支持多集合的对等差集运算。 """



# coding:utf-8

__all__ = ['List']

__metaclass__ = type

import collections
import redis.exceptions as redispy_exception

from ooredis.mix.key import (
    get_key_name_from_single_value,
    Key,
)
from ooredis.const import LEFTMOST, RIGHTMOST

# block time
INDEFINITELY = 0

# remove mode
REMOVE_ALL_ELEMENT_EQUAL_VALUE = 0

# redis blpop/brpop result index
VALUE = 1

# NOTE:
# redis的list是可变的(mutable)，没有继承MutableSequence主要是
# 考虑到python的列表是单向增长的(从左到右)，
# 而redis的列表是双向的，append、pop等和redis的表现不同。
class List(Key, collections.Sequence):
    """ 一个列表key对象，底层实现是redis的list类型。 """

    def __repr__(self):
        key_type = self.__class__.__name__.title()
        key_name = self.name
        key_values = list(self)
        return "{0} Key '{1}': {2}".format(key_type, key_name, key_values)

    def __len__(self):
        """ 返回列表中的元素个数。
        不存在的列表返回0。

        Time:
            O(1)

        Returns:
            int: 元素个数。

        Raises：
            TypeError: 尝试对非list类型的对象进行操作时抛出。
        """
        try:
            return self._client.llen(self.name)
        except redispy_exception.ResponseError:
            raise TypeError

    def __setitem__(self, key, value):
        """ 将列表指定位置的值设为value。

        暂时只支持对单个key赋值，像这样list[2] = 'hi'，
        不支持list[2:4] = ['go']这样的range赋值。

        Args:
            key: 指定列表的位置。
            value: 要设置的值。

        Time:
            O(1)

        Returns:
            None

        Raises:
            IndexError: key超出范围。
            TypeError: 当key对象不存在，
                       或对非list类型的对象进行操作时抛出。
            NotImplementedError: 当使用range赋值时抛出。
        """
        # TODO: 支持range赋值？

        # NOTE:
        # redis-py对不存的key，以及key超出范围都抛出ResponseError
        # ooredis里__getitem__和__setitem__对不存在的key都抛出IndexError
        # 主要是为了和python的list类型行为保持一致。
        if self.exists == False:
            raise IndexError

        if isinstance(key, slice):
            # 没有实现列表的range赋值，
            # 因为redis没有相应的函数。
            raise NotImplementedError
        elif key >= len(self):
            raise IndexError
        else:
            try:
                value = self._type_case.to_redis(value)
                self._client.lset(self.name, key, value)
            except redispy_exception.ResponseError:
                raise TypeError

    def __getitem__(self, key):
        """ 获取特定位置key上的列表元素。

        Args:
            key: 一个索引，或一个range。

        Time:
            获取表头或表尾为O(1)，其他为O(N)

        Returns:
            value: 列表元素。

        Raises:
            IndexError: 当key对象不存在，或key超出范围时抛出。
            TypeError: 尝试对非list类型的对象进行操作时抛出。
        """
        # WARNING: 注意redis的lrange函数和python的slice的差异，
        #          lrange的右边索引是闭区间，
        #          而slice的右边索引是开区间。
        #          这中间的细微差异让这个方法非常之难测试。
        #          除非你有详细且严谨的测试为基础，否则不要修改这个方法。
        #
        #          至于lindex函数，和python的list[key]行为相同，
        #          这里作为list[key]的get方法。
        try:
            if isinstance(key, slice):
                # 1.EDGE version(get key in server side):
                #
                #if key.start == key.stop != None: return []
                #
                #start = LEFTMOST if key.start == None else key.start
                #stop = RIGHTMOST if key.stop == None else key.stop-1
                #
                #return self._client.lrange(self.name, start, stop)
   
                # 2.SAFE version(get key in client side):
                value = self._client.lrange(self.name, LEFTMOST, RIGHTMOST)[key]
                return map(self._type_case.to_python, value)
            elif key >= len(self):
                raise IndexError
            else:
                value = self._client.lindex(self.name, key)
                return self._type_case.to_python(value)
        except redispy_exception.ResponseError:
            raise TypeError

    def __delitem__(self, key):
        """ 删除指定区间内的列表元素。

        目前只支持以下三种模式：
            del list[:]
            del list[x:]
            del list[:y]
        以下模式不支持：
            del list[index]
            del list[x:y]

        Args:
            key: 一个索引，或一个range。

        Time:
            O(N)

        Raises:
            IndexError: 当key对象不存在，或key超出范围时抛出。
            TypeError: 尝试对非list类型的对象进行操作时抛出。
        """
        # TODO: 想一个漂亮的办法(尽量不写元数据)解决
        #       del list[x:y] 和 del list[index]
        try:
            if isinstance(key, slice):
                # del list[:] --> del all item in list
                if key.start == key.stop == None:
                    # 当ltrim的start > stop时，清空整个列表 
                    self._client.ltrim(self.name, 1, 0)
                # del list[x:] --> keep list[0:x-1]
                elif key.stop == None:
                    self._client.ltrim(self.name, LEFTMOST, key.start-1) 
                # del list[:y] --> keep list[y:-1]
                elif key.start == None:
                    self._client.ltrim(self.name, key.stop, RIGHTMOST)
                # del list[x:y] --> keep list[0:x-1] and list[y:-1]
                else:
                    raise NotImplementedError
            else:
                raise NotImplementedError
        except redispy_exception.ResponseError:
            raise TypeError

    def remove(self, value):
        """ 移除列表中所有值等于value的元素。

        Args:
            value

        Time:
            O(N)

        Returns:
            被移除元素的数量。

        Raises:
            TypeError: 尝试对非list类型的对象进行操作时抛出。
        """
        try:
            value = self._type_case.to_redis(value)
            return self._client.lrem(self.name, value, REMOVE_ALL_ELEMENT_EQUAL_VALUE)
        except redispy_exception.ResponseError:
            raise TypeError

    # TODO: redis2.3将支持push多个值
    def lpush(self, value):
        """ 将value添加到表头。

        Args:
            value: 要添加的值。

        Time:
            O(1)

        Returns:
            None

        Raises:
            TypeError: 尝试对非list类型的对象进行操作时抛出。
        """
        try:
            value = self._type_case.to_redis(value)
            self._client.lpush(self.name, value)
        except redispy_exception.ResponseError:
            raise TypeError

    def lpop(self):
        """ 将表头元素弹出。

        Time:
            O(1)

        Returns:
            value: 表头元素。

        Raises:
            IndexError: 对空列表执行操作时抛出。
            TypeError: 尝试对非list类型的对象进行操作时抛出。
        """
        try:
            if len(self) == 0:
                raise IndexError
            value = self._client.lpop(self.name)
            return self._type_case.to_python(value)
        except redispy_exception.ResponseError:
            raise TypeError

    def blpop(self, timeout=INDEFINITELY):
        """ blpop是lpop命令的阻塞版本，
        当给定列表内没有任何元素可供弹出的时候，
        连接将被blpop命令阻塞，直到等待超时或发现可弹出元素为止。 


        Args:
            timeout: 等待的超时时间，0为一直等待。
                     默认为0。

        Returns:
            None: 当等到超时且没有元素被返回时。
            value: 列表不为空。

        Raises:
            TypeError: 尝试对非list类型的对象进行操作时抛出。
        """
        # redis原来的blpop可以对多个key进行blpop，这里只对单个key。
        # 因此，redis原来的返回值格式是(key, value)，
        # 而这里只返回value(原因同上)。
        try:
            result = self._client.blpop(self.name, timeout)

            value = result[VALUE] if result else None

            return self._type_case.to_python(value)
        except redispy_exception.ResponseError:
            raise TypeError

    # TODO: >= 2.4: 接受多个值
    def rpush(self, value):
        """ 将value添加到表尾。

        Args:
            value: 要添加到表尾的项。

        Time:
            O(1)

        Returns:
            None

        Raises:
            TypeError: 尝试对非list类型的对象进行操作时抛出。
        """
        try:
            value = self._type_case.to_redis(value) 
            self._client.rpush(self.name, value)
        except redispy_exception.ResponseError:
            raise TypeError

    def rpop(self):
        """ 将表尾元素弹出。

        Time:
            O(1)

        Returns:
            value: 表尾元素

        Raises:
            TypeError: 尝试对非list类型的对象进行操作时抛出。
            IndexError: 对空列表执行操作时抛出。
        """
        try:
            if len(self) == 0:
                raise IndexError
            value = self._client.rpop(self.name)
            return self._type_case.to_python(value)
        except redispy_exception.ResponseError:
            raise TypeError

    def brpop(self, timeout=INDEFINITELY ):
        """ brpop是rpop命令的阻塞版本，
        当给定列表内没有任何元素可供弹出的时候，
        连接将被brpop命令阻塞，直到等待超时或发现可弹出元素为止。 
        
        Args:
            timeout: 等待的超时时间，0为一直等待。
                     默认为0。

        Returns:
            None: 当等到超时且没有元素被返回时。
            value: 列表不为空。

        Raises:
            TypeError: 尝试对非list类型的对象进行操作时抛出。
        """
        # redis原来的brpop可以对多个key进行brpop，这里只对单个key。
        # 因此，redis原来的返回值格式是(key, value)，
        # 而这里只返回value(原因同上)。
        try:
            result = self._client.brpop(self.name, timeout)

            value = result[VALUE] if result else None

            return self._type_case.to_python(value)
        except redispy_exception.ResponseError:
            raise TypeError

    def rpoplpush(self, destination):
        """ 在一个原子时间内，完成以下三件事：
        1.对self进行rpop，
        2.将弹出的元素返回客户端，
        3.将该元素副本lpush到destination。

        Args:
            destination: 另一个列表，或列表名字。

        Time:
            O(1)

        Returns:
            None: 当列表为空时返回。
            value: 当列表不为空时，返回被弹出的元素。

        Raises:
            TypeError: 尝试对非list类型的对象进行操作时抛出。
        """
        try:
            destination = get_key_name_from_single_value(destination)
            value = self._client.rpoplpush(self.name, destination)
            return self._type_case.to_python(value)
        except redispy_exception.ResponseError:
            raise TypeError

    def brpoplpush(self, destination, timeout=INDEFINITELY):
        """ rpoplpush的阻塞版本。
        接受一个timeout参数作为阻塞超时时间。

        Args:
            destination: 另一个列表，或列表名字。
            timeout: 阻塞等待时间，默认为0。

        Time:
            O(1)

        Returns:
            None: 当列表为空时返回。
            value: 当列表不为空时，返回被弹出的元素。

        Raises:
            TypeError: 尝试对非list类型的对象进行操作时抛出。
        """
        try:
            destination = get_key_name_from_single_value(destination)
            value = self._client.brpoplpush(self.name, destination, timeout)
            return self._type_case.to_python(value)
        except redispy_exception.ResponseError:
            raise TypeError



#!/usr/bin/env python

from distutils.core import setup

setup(name='ooredis',
      version='1.0',
      author='huangz',
      author_email='huangz1990@gmail.com',
      url='https://github.com/huangz1990/ooredis',
      description='An redis python lib',
      long_description='',
      download_url='',
      packages=['ooredis', 'ooredis.mix',],
      classifiers = [
        'Topic :: Database',
        ],
     )



# coding: utf-8

import json
import pickle


class GenericTypeCase:
    """ 一般的类型转换类，可以处理str,unicode,int和float。 """

    @staticmethod
    def to_redis(value):
        # NOTE: unicode only
        if isinstance(value, basestring):
            value = unicode(value)
        return value

    @staticmethod
    def to_python(value):
        if value == None:
            return None

        # NOTE: 该转换的基础是在redis储存字符串
        #       的假定之下的，如果redis储存的不是
        #       字符串，该转换结果将是不准确的。
        assert(isinstance(value, basestring))

        try:
            return int(value)
        except:
            pass

        try:
            return float(value)
        except:
            pass

        try:
            return unicode(value)
        except:
            pass

        raise ValueError


class IntTypeCase:
    """ 将python值转换为int(long)类型。 """

    @staticmethod
    def to_redis(value):
        return int(value)

    @staticmethod
    def to_python(value):
        if value == None:
            return None
        return int(value)


class FloatTypeCase:
    """ 将python值转换为float类型。 """

    @staticmethod
    def to_redis(value):
        return float(value)

    @staticmethod
    def to_python(value):
        if value == None:
            return None
        return float(value)


class StringTypeCase:
    """ 将python值转换为字符串类型(unicode编码)。 """

    @staticmethod
    def to_redis(value):
        return unicode(value)

    @staticmethod
    def to_python(value):
        if value == None:
            return None
        return unicode(value)


class JsonTypeCase:
    """ 将python对象转化为JSON对象。 """

    @staticmethod
    def to_redis(value):
        return json.dumps(value)
    
    @staticmethod
    def to_python(value):
        if value == None:
            return None
        return json.loads(value)


class SerializeTypeCase:
    """ 将python对象序列化(使用pickle模块)。 """

    @staticmethod
    def to_redis(value):
        return pickle.dumps(value)

    @staticmethod
    def to_python(value):
        if value == None:
            return None
        return pickle.loads(value)
